package fileFormats;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.InvalidParameterException;
import java.util.HashMap;
import java.util.Map;
import java.util.Arrays;

public class XmlConverter implements Converter {

	//public static class BinaryXmlFileHelpers
		private int ReadValuePackedS32(InputStream stream) throws IOException
		{
			int val=0;
			int buff=0;
			for(int i=0;;i++)
			{
				buff=stream.read();
				val+=(buff&0x7f)<<(i*7);
				if((buff&0x80)==0)
					break;
			}
			return val;
		}
		
	protected Map<Integer, byte[]> textmap=new HashMap<Integer, byte[]>();

	
	private void parseData(InputStream input) throws IOException
	{
		int size = ReadValuePackedS32(input);
		
		byte[] buffer = new byte[size];
		input.read(buffer);
		int start=0;
		System.out.println(size);
		for(int i=0;i<buffer.length;i+=2)
		{
			if(buffer[i]+(buffer[i+1]<<8)==0)
			{
				textmap.put(start/2, Arrays.copyOfRange(buffer, start, i));
				start=i+2;
			}
		}
	}

	private class BinaryXmlNode
	{
		public byte[] _name;
		public byte[] _value;
		public Map<byte[], byte[]> _attributes;
		public BinaryXmlNode[] _children;
		//public List<BinaryXmlNode> Children;

		public void Read(InputStream input) throws IOException
		{
			_name=textmap.get(ReadValuePackedS32(input));

			int flags = input.read();

			// Value
			if ((flags & 1) != 0)
			{
				_value = textmap.get(ReadValuePackedS32(input));
			}

			// Attributes
			if ((flags & 2) != 0)
			{
				_attributes = new HashMap<byte[], byte[]>();
				int count = ReadValuePackedS32(input);
				for (int i = 0; i < count; i++)
				{
					_attributes.put(textmap.get(ReadValuePackedS32(input)), textmap.get(ReadValuePackedS32(input)));
				}
			}

			// Children
			if ((flags & 4) != 0)
			{
				int count = ReadValuePackedS32(input);
				_children=new BinaryXmlNode[count];
				for (int i = 0; i < count; i++)
				{
					BinaryXmlNode child = new BinaryXmlNode();
					child.Read(input);
					_children[i]=child;
				}
			}
		}
	}

	private BinaryXmlNode _root;
	
	public void Read(InputStream input, OutputStream output) throws IOException 
	{
		if (input.read() != 0x80)
		{
			throw new InvalidParameterException("not a binary XML file");
		}
 
		parseData(input);

		_root = new BinaryXmlNode();
		_root.Read(input);
		
		input.close();

		XMLWriter writer=new XMLWriter(output);
		writer.writeHead();
		writer.writeNode(_root, 0);

		output.flush();
		output.close();
	}
	public void Repair(InputStream input, InputStream input2, OutputStream output) throws IOException 
	{
		//do nothing
	}
	public boolean SupportRepair()
	{
		return false;
	}
	
	static byte[] SPACE = new byte[]{' ', 0};
	static byte[] ENTER = new byte[]{'\n', 0};
	static byte[] BEGIN = new byte[]{'<', 0};
	static byte[] END = new byte[]{'>', 0};
	static byte[] PER = new byte[]{'/', 0};
	static byte[] EQUAL = new byte[]{'=', 0};
	
	private class XMLWriter
	{
		private final OutputStream _output;
		public XMLWriter(OutputStream output)
		{
			_output = output;
		}
		
		public void writeHead() throws IOException
		{
	    	_output.write(0xfe);
	    	_output.write(0xff);
			byte[] header="<?xml version=\"1.0\"?>".getBytes();
			for(byte b:header)
			{
				_output.write(b);
				_output.write(0x00);
			}
			_output.write(ENTER);
			
    		_output.flush();
		}

		public void writeNode(BinaryXmlNode node, int depth) throws IOException
		{
			
	    	for(int i=0;i<depth*2;i++)
	    		_output.write(SPACE);
	    	
	    	_output.write(BEGIN);
	    	_output.write(node._name);
	    	
	    	if(node._attributes!=null)
	    	{
	    		for(Map.Entry<byte[], byte[]> e : node._attributes.entrySet())
	    		{
	    			_output.write(SPACE);
	    			_output.write(e.getKey());
	    			_output.write(EQUAL);
	    			_output.write(e.getValue());
	    		}
	    	}
	    	if(node._value==null&&(node._children==null||node._children.length==0))
	    	{
		    	_output.write(PER);
		    	_output.write(END);
		    	_output.write(ENTER);
		    	return;
	    	}

	    	_output.write(END);
	    	
	    	if(node._value!=null)
	    		_output.write(node._value);
	    	
	    	if(node._children!=null&&node._children.length>0)
	    	{
	    		_output.write(ENTER);
	    		for(BinaryXmlNode child:node._children)
	    			writeNode(child, depth+1);
	    		
		    	for(int i=0;i<depth*2;i++)
		    		_output.write(SPACE);
	    	}
	    	
    		_output.write(BEGIN);
    		_output.write(PER);
    		_output.write(node._name);
    		_output.write(END);
    		_output.write(ENTER);		   
    		
    		_output.flush();
	    }
		
	}

}
